# frozen_string_literal: true

# The Puppet Extensions Module.
#
# Submodules of this module should be named after the publisher (e.g.
# 'user' part of a Puppet Module name). You should also add the module
# name to avoid further conflicts in the same namespace. So the
# structure should be
# `lib/puppet_x/<namespace>/<module_name>/<extension.rb>`.
# A Puppet Extension for 'puppetlabs-powershell' would live at
# `lib/puppet_x/puppetlabs/powershell/<extension.rb>`.
#
# @api public
#
module PuppetX
end
