# @summary Notify to demonstrate users::common in catalog
#
# A common Class that all examples should include
#
# @example
#   include users::common
class users::common {
  notify { 'Adding users::common': }
}
